# jps2sm.py is a python script that will automatically gather data from JPS from a given group url, release url,
# or a user's uploaded / seeding torrents and iterate through them and upload them to SM.

# Catch python2 being run here to get a relatively graceful error message, rather than a syntax error later on which causes confusion.
from typing import List, Any

error_x, *error_y = 1, 2, 3, 4  # jps2sm requires requires python3.8, a SyntaxError here means you are running it in python2!

# Catch python < 3.8 being run here to get a relatively graceful error message, rather than a syntax error later on which causes confusion.
print(walrus := "", end='')  # jps2sm requires python3.8, a SyntaxError here means you are running it in python <= 3.7!

# Standard version check that for now it pretty useless
import sys

if sys.version_info < (3, 8):
    print("Error: jps2sm requires python 3.8 to run.", file=sys.stderr)
    exit(1)

# Standard library packages
import re
import os
import datetime
import collections
import configparser
import json
import io

# Third-party packages
from bs4 import BeautifulSoup
import humanfriendly
from pathlib import Path
from loguru import logger
from datasize import DataSize

# jps2sm modules
from jps2sm.get_data import GetGroupData, get_release_data, GetJPSUser, GetSMUser
from jps2sm.save_data import save_sm_html_debug_output, download_sm_uploaded_torrents, download_sm_torrent, get_jps_torrent, download_jps_torrent
from jps2sm.batch import batch_mode
from jps2sm.utils import fatal_error, GetConfig, GetArgs, decide_duplicate
from jps2sm.myloginsession import jpopsuki, sugoimusic
from jps2sm.constants import Categories, VideoOptions
from jps2sm.mediainfo import get_mediainfo
from jps2sm.validation import decide_music_performance, get_alternate_fansub_category_id, validate_jps_video_data, validate_jps_bitrate, \
    decide_exc_filter, decide_ep


def detect_display_swapped_names(userid):
    """
    Detect if the user has original (Japanese/Chinese characters) names shown for Artists and Torrent groups in their torrents.php views
    :param userid:
    :return: True if enabled (bad) or False if disabled (OK)
    """
    user_profile_page = jpopsuki(f"https://jpopsuki.eu/user.php?action=edit&userid={userid}")
    soup = BeautifulSoup(user_profile_page.text, 'html5lib')
    user_form = str(soup.select('#content .thin #userform'))

    # We do both string matches to be extra safe due to the havoc it causes if the user has original characters in torrent lists switched on
    good_setting = user_form.find('<input id="browsejp" name="browsejp" type="checkbox"/>')
    bad_setting = user_form.find('<input checked="checked" id="browsejp" name="browsejp" type="checkbox"/>')

    if (good_setting != -1) and (bad_setting == -1):
        # OK!
        return False
    else:
        # Not OK!
        return True


def setorigartist(artist, origartist):
    """
    Set an artist's original artist with the string origartist, currently used for contrib artists
    # TODO Consider using this for main orig artist

    :param artist: string: String of the artist that needs it's original artist set
    :param origartist: string: Original artist
    """
    if origartist == "":  # If there is no origartist at JPS do not bother trying to set it here
        return

    SMartistpage = sugoimusic(f"https://sugoimusic.me/artist.php?artistname={artist}")

    if re.findall("Your search did not match anything", SMartistpage.text):
        logger.debug(f"Artist {artist} does not yet exist at SM so origartist cannot be set")
        return

    soup = BeautifulSoup(SMartistpage.text, 'html5lib')
    linkbox = str(soup.select('#content .thin .header .linkbox'))
    artistid = re.findall(r'href="artist\.php\?action=edit&amp;artistid=([0-9]+)"', linkbox)[0]

    sm_user = GetSMUser()

    data = {
        'action': 'edit',
        'auth': sm_user.auth_key(),
        'artistid': artistid,
        'name_jp': origartist
    }

    SMeditartistpage = sugoimusic(f'https://sugoimusic.me/artist.php?artistname={artist}', 'post', data)
    logger.debug(f'Set artist {artist} original artist to {origartist}')


def uploadtorrent(jps_torrent_object, torrentgroupdata, **uploaddata):
    """
    Prepare POST data for the SM upload, performs additional validation, reports errors and performs the actual upload to
    SM whilst saving the html result to investigate any errors if they are not reported correctly.

    :param jps_torrent_object: bytes: BytesIO object of the JPS torrent
    :param groupid: groupid to upload to - allows to upload torrents to the same group
    :param uploaddata: dict of collated / validated release data from collate()
    """
    config = GetConfig()
    args = GetArgs()
    uploadurl = 'https://sugoimusic.me/upload.php'
    languages = ('Japanese', 'English', 'Korean', 'Chinese', 'Vietnamese')

    if torrentgroupdata.date is None:  # If release date cannot be derived use upload date
        date = uploaddata['uploaddate']
    else:
        date = torrentgroupdata.date

    data = {
        'submit': 'true',
        'title': torrentgroupdata.title,
        'year': date,
        'tags': torrentgroupdata.tagsall,
        'album_desc': torrentgroupdata.groupdescription,
        # 'release_desc': releasedescription
    }

    sm_user = GetSMUser()

    if not args.parsed.dryrun:
        data['auth'] = sm_user.auth_key()

    logger.debug(uploaddata)

    # TODO Most of this can be in getmediainfo()
    if args.parsed.mediainfo:
        try:
            data['mediainfo'], releasedatamediainfo = get_mediainfo(jps_torrent_object, uploaddata['media'], config.media_roots)
            data.update(releasedatamediainfo)
            if 'duration' in data.keys() and data['duration'] > 1:
                duration_friendly_format = humanfriendly.format_timespan(datetime.timedelta(seconds=int(data['duration'] / 1000)))
                data['album_desc'] += f"\n\nDuration: {duration_friendly_format} - {str(data['duration'])}ms"
        except Exception as mediainfo_exc:
            if str(mediainfo_exc).startswith('Mediainfo error - file/directory not found'):
                pass
            if str(mediainfo_exc).startswith('Mediainfo error - unable to extract what appears to be a Bluray disc:'):
                pass
            if torrentgroupdata.category in Categories.Video:
                raise
            else:
                logger.debug(f'Skipping exception on mediainfo failing as {torrentgroupdata.title} is not a Video category.')

    if torrentgroupdata.category not in Categories.NonReleaseData:
        data['media'] = uploaddata['media']
        if 'audioformat' not in data.keys():  # If not supplied by getmediainfo() use audioformat guessed by collate()
            data['audioformat'] = uploaddata['audioformat']

    if torrentgroupdata.imagelink is not None:
        data['image'] = torrentgroupdata.imagelink

    if uploaddata['videotorrent']:
        if torrentgroupdata.category == "DVD" and uploaddata['media'] == 'Bluray':
            data['type'] = Categories.JPStoSM['Bluray']  # JPS has no Bluray category
        if uploaddata['categorystatus'] == 'bad':  # Need to set a correct category
            if uploaddata['media'] == 'Bluray':
                data['type'] = Categories.JPStoSM['Bluray']
            else:  # Still need to change the category to something, if not a Bluray then even if it is not a DVD the most sensible category is DVD in a music torrent group
                data['type'] = Categories.JPStoSM['DVD']
        if torrentgroupdata.category == "TV-Music" and args.parsed.mediainfo:
            data['type'] = Categories.SM[decide_music_performance(torrentgroupdata.artist, data['multiplefiles'], data['duration'])]

        # If not supplied by getmediainfo() use codec found by collate()
        if 'codec' not in data.keys():
            data['codec'] = uploaddata['codec']

        # If not supplied by getmediainfo() try to detect resolution by searching the group description for resolutions
        if 'ressel' not in data.keys():
            foundresolutions720 = re.findall('1080 ?x ?720', torrentgroupdata.groupdescription)
            foundresolutions1080 = re.findall('1920 ?x ?1080', torrentgroupdata.groupdescription)
            if len(foundresolutions720) != 0:
                data['ressel'] = "720p"
            elif len(foundresolutions1080) != 0:
                data['ressel'] = "1080p"
            for resolution in VideoOptions.resolutions:  # Now set more specific resolutions if they are present
                if resolution in torrentgroupdata.groupdescription:  # If we can see the resolution in the group description then set it
                    data['ressel'] = resolution
                else:
                    data['ressel'] = 'CHANGEME'

        # If not supplied by getmediainfo() use container found by collate()
        if 'container' not in data.keys():
            data['container'] = uploaddata['container']

        data['sub'] = 'NoSubs'  # assumed default
        data['lang'] = 'CHANGEME'
        for language in languages:  # If we have a language tag, set the language field
            if language.lower() in torrentgroupdata.tagsall:
                data['lang'] = language
    elif torrentgroupdata.category in Categories.Music:
        data['bitrate'] = uploaddata['bitrate']

    if 'remastertitle' in uploaddata.keys():
        data['remaster'] = 'remaster'
        data['remastertitle'] = uploaddata['remastertitle']
    if 'remasteryear' in uploaddata.keys():
        data['remaster'] = 'remaster'
        data['remasteryear'] = uploaddata['remasteryear']

    # Non-BR/DVD/TV-* category validation
    # TODO Move this to a def
    if torrentgroupdata.category == "Fansubs":
        data['type'] = get_alternate_fansub_category_id(torrentgroupdata.artist, torrentgroupdata.title)  # Title just for user
        data['sub'] = 'Hardsubs'  # We have subtitles! Subs in JPS FanSubs are usually Hardsubs so guess as this
        # TODO: Use torrent library to look for sub/srt files
    elif torrentgroupdata.category == "Album":  # Ascertain if upload is EP
        data['type'] = Categories.JPStoSM[decide_ep(jps_torrent_object, uploaddata)]

    if 'type' not in data.keys():  # Set default value after all validation has been done
        data['type'] = Categories.JPStoSM[torrentgroupdata.category]

    # Now that all Category validation is complete decide if we should strip some mediainfo data
    mediainfo_non_resolution = ('container', 'mediainfo')
    mediainfo_resolution = ('ressel', 'resolution')
    if args.parsed.mediainfo and data['type'] in Categories.SM_StripAllMediainfo:
        for field in (mediainfo_non_resolution + mediainfo_resolution):
            data.pop(field, None)
    elif args.parsed.mediainfo and data['type'] == Categories.SM_StripAllMediainfoExcResolution:
        for field in mediainfo_non_resolution:
            data.pop(field, None)

    try:
        data['artist_jp'] = torrentgroupdata.originalartist
        data['title_jp'] = torrentgroupdata.originaltitle
    except AttributeError:  # If no originalchars do nothing
        pass

    try:
        contribartistsenglish = []
        for artist, origartist in torrentgroupdata.contribartists.items():
            contribartistsenglish.append(artist)
        data['contrib_artists[]'] = contribartistsenglish
    except AttributeError:  # If no contrib artists do nothing
        pass

    if "V.A." in torrentgroupdata.artist:  # At JPS Various Artists torrents have their artists as contrib artists
        del data['contrib_artists[]']  # Error if null as if there is a V.A. torrent group with no contrib artists something is wrong
        data['idols[]'] = contribartistsenglish
        logger.debug(f'Various Artists torrent, setting main artists to {contribartistsenglish}')
    else:
        data['idols[]'] = torrentgroupdata.artist  # Set the artist normally

    jps_torrent_object.seek(0)

    postDataFiles = {
        # We need to specify a filename  now we are using BytesIO and SM will validate files without a .torrent extension
        'file_input': ('blah.torrent', jps_torrent_object)
    }

    if args.parsed.dryrun or args.parsed.debug:
        dataexcmediainfo = {x: data[x] for x in data if x not in 'mediainfo'}
        dataexcmediainfo['auth'] = '<scrubbed>'
        logger.info(json.dumps(dataexcmediainfo, indent=2))  # Mediainfo shows too much data
    if not args.parsed.dryrun:
        SMres = sugoimusic(uploadurl, "post", data, postDataFiles)

        SMerrorTorrent = re.findall('red; text-align: center;">(.*)</p>', SMres.text)
        if SMerrorTorrent:
            dupe = re.findall('torrentid=([0-9]+)">The exact same torrent file already exists on the site!</a>$', SMerrorTorrent[0])
            if dupe:
                if not config.skip_dupes:
                    dupe_file_path = download_sm_torrent(dupe[0], torrentgroupdata.artist, torrentgroupdata.title)
                    logger.warning(
                        f'This torrent already exists on SugoiMusic - https://sugoimusic.me/torrents.php?torrentid={dupe[0]} '
                        f'The .torrent has been downloaded with name "{dupe_file_path}"'
                    )
                dupe_error_msg = f'The exact same torrent file already exists on the site! See: https://sugoimusic.me/torrents.php?torrentid={dupe[0]} JPS torrent id: {uploaddata["jpstorrentid"]}'
                logger.info(dupe_error_msg)
                raise Exception(dupe_error_msg)
            else:
                logger.error(SMerrorTorrent[0])
                raise Exception(SMerrorTorrent[0])

        SMerrorLogon = re.findall('<p>Invalid (.*)</p>', SMres.text)
        if SMerrorLogon:
            raise Exception(f'Invalid {SMerrorLogon[0]}')

        html_debug_output_path = save_sm_html_debug_output(SMres.text, torrentgroupdata, uploaddata['jpstorrentid'])

        groupid = re.findall('<input type="hidden" name="groupid" value="([0-9]+)" />', SMres.text)
        if not groupid:
            # Find groupid if private torrent warning
            groupid = re.findall(
                r'Your torrent has been uploaded; however, you must download your torrent from <a href="torrents\.php\?id=([0-9]+)">here</a>',
                SMres.text)
            if not groupid:
                unknown_error_msg = f'Cannot find groupid in SM response - there was probably an unknown error. See {html_debug_output_path} for potential errors'
                raise RuntimeError(unknown_error_msg)

        if groupid:
            logger.info(f'Torrent uploaded successfully as groupid {groupid[0]}  See https://sugoimusic.me/torrents.php?id={groupid[0]}')


def collate(torrentids, torrentgroupdata, max_size=None, scrape_only=False):
    """
    Collate and validate data ready for upload to SM

    Validate and process dict supplied by getreleasedata() with format, bitrate, media, container, codec, and remaster data to extract
    all available data from JPS
    Perform validation on some fields
    Download JPS torrent
    Apply filters via decide_exc_filter()
    Send data to uploadtorrent()
    Send data to setorigartists()

    :param torrentids: list of JPS torrentids to be processed
            Always a single torrentid unless specifying a group url in --url mode
    :param torrentgroupdata: dictionary with torrent group data from getgroupdata[]
    :param max_size: str: Maximum size with unit specified, currently only used by recent mode
    :param scrape_only: bool: Only download JPS torrents, do not upload to SM
    """
    config = GetConfig()
    args = GetArgs()

    jps_torrent_downloaded_count = sm_torrent_uploaded_count = skipped_max_size = skipped_low_seeders = skipped_exc_filter = skipped_dupe_torrent_hash = 0
    dupe_jps_ids = []
    dupe_sm_ids = []

    for jps_torrent_id, release_data in get_release_data(torrentids, torrentgroupdata.torrent_table, torrentgroupdata.date).items():

        logger.info(f'Now processing: {jps_torrent_id} {release_data}')

        if max_size:
            good_jps_format = release_data['size_units'][:1] + "i" + release_data['size_units'][1:]  # JPS uses 'GB' when it means 'GiB' etc.
            jps_torrent_size_in_bytes = DataSize(release_data['size_no_units'] + good_jps_format)
            cfg_max_size_in_bytes = DataSize(max_size)
            if jps_torrent_size_in_bytes > cfg_max_size_in_bytes:
                skipped_max_size += 1
                logger.debug(f"Skipping as torrent is > {max_size}")
                continue

        if int(release_data['seeders']) < config.jps_min_seeders:
            logger.debug(f'Skipping as torrent has < {config.jps_min_seeders} seeder(s)')
            skipped_low_seeders += 1
            continue

        slash_data = release_data['slashdata']
        uploaddatestr = release_data['uploaddate']
        release_data_collated = {'jpstorrentid': jps_torrent_id}
        remaster_data = False  # Set default

        # JPS uses the audioformat field (represented as slash_data[0] here) for containers and codecs in video torrents,
        # and when combined with VideoMedias we can perform VideoTorrent detection.
        if slash_data[0] in VideoOptions.badformats and slash_data[1] in VideoOptions.VideoMedias:
            # VideoCategory torrent, this also detects VideoCategories in a non-VC group
            # container / media
            release_data_collated['videotorrent'] = True  # For processing by uploadtorrent()
            release_data_collated['categorystatus'] = "good"

            #videoreleasedatavalidated = validate_jps_video_data(slash_data, release_data_collated['categorystatus'])
            for field, data in validate_jps_video_data(slash_data, release_data_collated['categorystatus']).items():
                release_data_collated[field] = data

            if len(slash_data) == 3:  # Remastered
                remaster_data = slash_data[2]

        elif slash_data[0] in VideoOptions.badformats and slash_data[2] in VideoOptions.VideoMedias:
            # Video torrent mistakenly uploaded as an Album/Single
            # container / 'bitrate' / media   Bitrate is meaningless, users usually select Lossless
            release_data_collated['videotorrent'] = True  # For processing by uploadtorrent()
            release_data_collated['categorystatus'] = "bad"

            videoreleasedatavalidated = validate_jps_video_data(slash_data, release_data_collated['categorystatus'])
            for field, data in videoreleasedatavalidated.items():
                release_data_collated[field] = data

            if len(slash_data) == 4:  # Remastered
                remaster_data = slash_data[3]

        elif torrentgroupdata.category in Categories.Music:  # Music torrent
            # format / bitrate / media
            release_data_collated['videotorrent'] = False
            release_data_collated['categorystatus'] = "good"

            release_data_collated['media'] = slash_data[2]
            release_data_collated['audioformat'] = slash_data[0]
            release_data_collated['bitrate'] = validate_jps_bitrate(slash_data[1])

            if decide_exc_filter(release_data_collated['audioformat'], release_data_collated['media'], slash_data):
                skipped_exc_filter += 1
                continue

            if len(slash_data) == 4:  # Remastered
                remaster_data = slash_data[3]

        elif torrentgroupdata.category in Categories.Video:  # Probably Music in a VC group
            # format / media
            release_data_collated['videotorrent'] = False
            release_data_collated['categorystatus'] = "bad"

            release_data_collated['audioformat'] = slash_data[0]
            release_data_collated['media'] = slash_data[1]

            if decide_exc_filter(release_data_collated['audioformat'], release_data_collated['media'], slash_data):
                skipped_exc_filter += 1
                continue

            if len(slash_data) == 3:  # Remastered
                remaster_data = slash_data[2]

        elif torrentgroupdata.category in Categories.NonReleaseData:  # Pictures or Misc Category torrents
            release_data_collated['videotorrent'] = False
            release_data_collated['categorystatus'] = "good"

        else:  # We should never reach here
            logger.error('Could not handle release data')
            raise RuntimeError('Could not handle release data')

        if remaster_data:
            try:
                remaster_text = re.findall('(.*) - (.*)$', remaster_data)[0]
                release_data_collated['remastertitle'] = remaster_text[0]
                remaster_year = remaster_text[1]
            except IndexError:  # Torrent is remastered and only has year set
                remaster_year = remaster_data  # The whole string is just the year

            # Year is mandatory on JPS so most remastered releases have current year set as year. This looks ugly on SM (and JPS) so if the
            # year is the groupdata['year'] we will not set it.
            year = re.findall('([0-9]{4})(?:.*)', torrentgroupdata.date)[0]
            if year != remaster_year:
                release_data_collated['remasteryear'] = remaster_year

        if 'WEB' in slash_data:  # Media validation
            release_data_collated['media'] = 'Web'
        elif 'Blu-Ray' in slash_data:
            release_data_collated['media'] = 'Bluray'  # JPS may actually be calling it the correct official name, but modern usage differs.

        # uploadtorrent() will use the upload date as release date if the torrent has no release date, usually for
        # Picture Category torrents and some TV-Variety.
        release_data_collated['uploaddate'] = datetime.datetime.strptime(uploaddatestr, '%b %d %Y, %H:%M').strftime('%Y%m%d')

        jps_torrent_file = get_jps_torrent(jps_torrent_id, torrentgroupdata)
        jps_torrent_object = io.BytesIO(jps_torrent_file.content)  # Keep file in memory as it could be processed and deleted by a torrent client

        dupe_sugoimusic_torrent_id = decide_duplicate(jps_torrent_object)

        if dupe_sugoimusic_torrent_id and config.skip_dupes:
            logger.debug(f'Not downloading JPS torrent {jps_torrent_id} as it is a duplicate on SM as torrent {dupe_sugoimusic_torrent_id}'
                         f' and SkipDuplicates is true in cfg.')
            skipped_dupe_torrent_hash += 1
        else:
            download_jps_torrent(jps_torrent_file, torrentgroupdata, slash_data)
            logger.debug(f'downloaded jps torrent {jps_torrent_id}')
            jps_torrent_downloaded_count += 1

        if scrape_only:
            continue

        if not args.parsed.dryrun and dupe_sugoimusic_torrent_id:
            if not config.skip_dupes:
                dupe_file_path = download_sm_torrent(dupe_sugoimusic_torrent_id, torrentgroupdata.artist, torrentgroupdata.title)
                # torrentgroupdata.artist and torrentgroupdata.title is just to generate a pretty filename
                logger.warning(
                    f'This torrent already exists on SugoiMusic - https://sugoimusic.me/torrents.php?torrentid={dupe_sugoimusic_torrent_id} '
                    f'The .torrent has been downloaded with name "{dupe_file_path}"'
                )
            dupe_error_msg = f'The exact same torrent file already exists on the site! See: https://sugoimusic.me/torrents.php?torrentid={dupe_sugoimusic_torrent_id} JPS torrent id: {jps_torrent_id}'
            logger.warning(dupe_error_msg)
            dupe_jps_ids.append(int(jps_torrent_id))
            dupe_sm_ids.append(int(dupe_sugoimusic_torrent_id))
            continue

        # Upload torrent to SM
        uploadtorrent(jps_torrent_object, torrentgroupdata, **release_data_collated)

        sm_torrent_uploaded_count += 1

    if not args.parsed.dryrun:
        # Add original artists for contrib artists
        if torrentgroupdata.contribartists:
            for artist, origartist in torrentgroupdata.contribartists.items():
                # For every artist, go to its artist page to get artist ID, then use this to go to artist.php?action=edit with the orig artist
                try:
                    setorigartist(artist, origartist)
                except IndexError:  # Do not let a setorigartist error affect stats
                    logger.debug(f'Error in setting artist {artist} origartist {origartist}')
                    pass

    collate_torrent_info = {
        'jps_torrents_downloaded_count': jps_torrent_downloaded_count,
        'sm_torrents_uploaded_count': sm_torrent_uploaded_count,  # For use by download_sm_uploaded_torrents() or statisics when in a batch_mode
        'skipped_torrents_max_size': skipped_max_size,
        'skipped_torrents_low_seeders': skipped_low_seeders,
        'skipped_torrents_exc_filter': skipped_exc_filter,
        'skipped_torrents_duplicate': skipped_dupe_torrent_hash,
        'dupe_jps_ids': dupe_jps_ids,
        'dupe_sm_ids': dupe_sm_ids,
    }

    return collate_torrent_info


def main():
    args = GetArgs()

    if args.parsed.debug:
        stderr_log_level = "DEBUG"
    else:
        stderr_log_level = "INFO"
    logger.remove()
    logger.add(sys.stderr, format="<lvl>{message}</>", level=stderr_log_level)
    logger.add("jps2sm.log", level="DEBUG", rotation="1 MB")

    if not args.parsed.debug:
        sys.tracebacklimit = 0

    if args.parsed.mediainfo:
        config = GetConfig()
        try:
            for media_dir in config.media_roots:
                if not os.path.exists(media_dir):
                    fatal_error(f'Error: Media directory {media_dir} does not exist. Check your configuration in jps2sm.cfg.')
                if not os.path.isdir(media_dir):
                    fatal_error(f'Error: Media directory {media_dir} is a file and not a directory. Check your configuration in jps2sm.cfg.')
        except configparser.NoSectionError:
            fatal_error('Error: --mediainfo requires you to configure MediaDirectories in jps2sm.cfg for mediainfo to find your file(s).')

    jps_user = GetJPSUser()
    jps_user_id = jps_user.user_id()
    logger.debug(f"JPopsuki user id is {jps_user_id}")

    if detect_display_swapped_names(jps_user_id):
        fatal_error("Error: 'Display original Artist/Album titles' is enabled in your JPS user profile. This must be disabled for jps2sm to run.")

    if args.parsed.torrentid:
        non_batch_upload(jps_torrent_id=args.parsed.torrentid, dry_run=args.parsed.dryrun, wait_for_jps_dl=args.parsed.wait_for_jps_dl)
        return

    if args.parsed.urls:
        non_batch_upload(jps_urls=args.parsed.urls, dry_run=args.parsed.dryrun, wait_for_jps_dl=args.parsed.wait_for_jps_dl)
        return

    if args.batch_modes == 1:
        batch_mode_user = args.parsed.batchuser or jps_user_id
        batch_mode(mode=args.batch_mode, user=batch_mode_user, start=args.parsed.batchstart,
                   end=args.parsed.batchend, sort=args.parsed.batchsort, order=args.parsed.batchsortorder)
    else:
        # If we reach here something has gone very wrong with parsing args
        raise RuntimeError('Argument handling error')


def non_batch_upload(jps_torrent_id=None, jps_urls=None, dry_run=None, wait_for_jps_dl=False):
    """
    Perform simple non-batch upload to SM with either a single jps_torrent_id from --torrentid or a string of --urls
    """

    if jps_torrent_id and jps_urls:
        raise RuntimeError('Expected either jps_torrent_id OR jps_url')

    if jps_torrent_id:
        jps_group_data = GetGroupData(f"https://jpopsuki.eu/torrents.php?torrentid={jps_torrent_id}")
        # This is a hack to ensure the get_release_data() works as it is designed to only parse a list of strs
        jps_torrent_ids = [str(jps_torrent_id)]
    elif jps_urls:
        jps_group_data = GetGroupData(jps_urls)
        jps_torrent_ids = re.findall('torrentid=([0-9]+)', jps_urls)
    else:
        raise RuntimeError('Expected either a jps_torrent_id or a jps_url')

    if wait_for_jps_dl:
        pre_torrent_info = collate(torrentids=jps_torrent_ids, torrentgroupdata=jps_group_data, scrape_only=True)
        input('When these files have been downloaded press enter to continue...')

    torrent_info = collate(torrentids=jps_torrent_ids, torrentgroupdata=jps_group_data)
    if not dry_run:
        download_sm_uploaded_torrents(torrent_info['sm_torrents_uploaded_count'], jps_group_data.artist, jps_group_data.title)


if __name__ == "__main__":
    main()



# Standard library packages
import re

from jps2sm.myloginsession import jpopsuki
from jps2sm.constants import Categories, VideoOptions
from jps2sm.utils import GetArgs

# Third-party packages
from bs4 import BeautifulSoup
import torrent_parser as tp
from loguru import logger


def decide_music_performance(artists, multiplefiles, duration):
    """
    Return if upload should be a Music Performance or not
    A music performance is a cut from a Music TV show and is 25 mins or less long and therefore also not a TV Show artist

    We assume we are being called if Cat = TV Music

    :return:  str: 'Music Performance' or 'TV Music'
    """
    if multiplefiles is True or duration > 1500000:  # 1 500 000 ms = 25 mins
        return 'TV Music'
    else:  # Single file that is < 25 mins, decide if Music Performance
        if len(artists) > 1:  # Multiple artists
            logger.debug('Upload is a Music Performance as it has derived multiple artists and is 25 mins or less')
            return 'Music Performance'  # JPS TV Show artists never have multiple artists
        JPSartistpage = jpopsuki(f"https://jpopsuki.eu/artist.php?name={artists[0]}")
        soup = BeautifulSoup(JPSartistpage.text, 'html5lib')
        categoriesbox = str(soup.select('#content .thin .main_column .box.center'))
        categories = re.findall(r'\[(.+)\]', categoriesbox)
        if any({*Categories.NonTVCategories} & {*categories}):  # Exclude any TV Shows for being mislabeled as Music Performance
            logger.debug('Upload is a Music Performance as it is 25 mins or less and not a TV Show')
            return 'Music Performance'
        else:
            logger.debug('Upload is not a Music Performance')
            return 'TV Music'


def get_alternate_fansub_category_id(artist, group_name):
    """
    Attempts to detect the actual category for JPS Fansubs category torrents and if not ask the user to select an alternate category.
    If it is a TV show, this TV show category type is detected and returned, else query the user from a list of potential categories.

    :param artist: str artist name
    :return: int alternative category ID based on Categories.SM()
    """
    JPSartistpage = jpopsuki(f"https://jpopsuki.eu/artist.php?name={artist}")
    soup = BeautifulSoup(JPSartistpage.text, 'html5lib')
    categoriesbox = str(soup.select('#content .thin .main_column .box.center'))
    categories = re.findall(r'\[(.+)\]', categoriesbox)

    if not any({*Categories.NonTVCategories} & {*categories}) and " ".join(categories).count('TV-') == 1:
        # Artist has no music and only 1 TV Category, artist is a TV show and we can auto detect the category for FanSub releases
        autodetectcategory = re.findall(r'(TV-(?:[^ ]+))', " ".join(categories))[0]
        logger.debug(f'Autodetected SM category {autodetectcategory} for JPS Fansubs torrent')
        return autodetectcategory
    else:  # Cannot autodetect
        AlternateFanSubCategoriesIDs = (5, 6, 7, 8, 9, 11)  # Matches indices in Categories()
        logger.warning(f'Cannot auto-detect correct category for torrent group {group_name}.')
        print('Select Category:')
        option = 1
        optionlookup = {}
        for alternativefansubcategoryid in AlternateFanSubCategoriesIDs:
            for cat, catid in Categories.SM.items():
                if alternativefansubcategoryid == catid:
                    print(f'({option}) {cat}')
                    optionlookup[option] = alternativefansubcategoryid
                    option += 1
        alternatecategoryoption = input('Choose alternate category or press ENTER to skip: ')
        if alternatecategoryoption == "":
            logger.error('No alternate Fansubs category chosen.')
            return "Fansubs"  # Allow upload to fail
        else:
            category = optionlookup[int(alternatecategoryoption)]
            logger.info(f'Alternate Fansubs category {category} chosen')
            return category


def validate_jps_video_data(releasedata, categorystatus):
    """
    Validate and process dict supplied by getreleasedata() via collate() to extract all available data
    from JPS for video torrents, whilst handling weird cases where VideoTorrent is uploaded as a Music category

    :param releasedata:
    :param categorystatus: str: good or bad. good for correct category assigned and bad if this is a Music Torrent
    mistakenly uploaded as a non-VC category!
    :return: releasedataout{} validated container, codec, media, audioformat
    """
    releasedataout = {}
    # JPS uses the audioformat field (represented as releasedata[0] here) for containers and codecs in video torrents

    # If a known container is used as audioformat set it as the container on SM
    if releasedata[0] in VideoOptions.badcontainers:
        releasedataout['container'] = releasedata[0]
    else:
        releasedataout['container'] = 'CHANGEME'
    # If a known codec is used as audioformat set it as the codec on SM
    if releasedata[0] in VideoOptions.badcodecs:
        if releasedata[0] == "MPEG2":  # JPS uses 'MPEG2' for codec instead of the correct 'MPEG-2'
            releasedataout['codec'] = "MPEG-2"
        else:
            releasedataout['codec'] = releasedata[0]
    else:
        releasedataout['codec'] = 'CHANGEME'  # assume default

    if categorystatus == "good":
        releasedataout['media'] = releasedata[1]
    else:
        releasedataout['media'] = releasedata[2]

    if releasedata[0] == 'AAC':  # For video torrents, the only correct audioformat in JPS is AAC
        releasedataout['audioformat'] = "AAC"
    else:
        releasedataout['audioformat'] = "CHANGEME"

    return releasedataout


def validate_jps_bitrate(jps_bitrate):
    """
    Validate JPS bad bitrates to sensible bitrates ready for upload to SM

    :param jps_bitrate:
    :return: sm_bitrate
    """

    bitrates = {
        "Hi-Res 96/24": "24bit Lossless 96kHz",
        "24bit/48kHz": "24bit Lossless 48kHz",
        "Hi-Res": "24bit Lossless",
        "24/44.1": "24bit Lossless",
        "Hi-Res 48/24": "24bit Lossless 48kHz",
        "Hi-Res 96-24": "24bit Lossless 96kHz",
        "24bit/96kHz": "24bit Lossless 96kHz",
        "24bit/48Khz": "24bit Lossless 48kHz",
        "24bit/96Khz": "24bit Lossless 96kHz",
        "24bit/48khz": "24bit Lossless 48kHz",
        "24/48": "24bit Lossless 48kHz",
        "24/96": "24bit Lossless 96kHz",
        "Hi-Res Lossless": "24bit Lossless",
        "160": "Other",
        "Variable": "Other",
        "320 (VBR)": "Other",
        "Scans": "",
        "Booklet": "",
        "1080p": "",
        "720p": "",
        "256 (VBR)": "APS (VBR)",
        "155": "Other"
    }

    sm_bitrate = jps_bitrate  # Default is to leave bitrate alone if not mentioned here, such as bitrates that are OK on both JPS and SM
    for old, new in bitrates.items():
        if jps_bitrate == old:
            sm_bitrate = new

    return sm_bitrate


def decide_exc_filter(audioformat, media, releasedata):
    """
    Implement audioformat and media exclusion filters
    :return: boolean: True or False
    """
    args = GetArgs()

    if audioformat == args.parsed.excaudioformat:
        logger.info(f'Excluding {releasedata} as exclude audioformat {args.parsed.excaudioformat} is set')
        return True
    elif media == args.parsed.excmedia:
        logger.info(f'Excluding {releasedata} as exclude media {args.parsed.excmedia} is set')
        return True

    return False


def decide_ep(jps_torrent_object, uploaddata):
    """
    Return if Album upload should be an EP or not.
    EPs are considered to have < 7 tracks, excluding off-vocals and uploaded to JPS as an Album

    We assume we are being called only if Cat = Album

    :param jps_torrent_object: bytes: BytesIO object of the JPS torrent
    :param uploaddata:
    :return: str: 'EP' or 'Album'
    """

    if uploaddata['media'].lower() == 'bluray' or uploaddata['media'].lower() == 'dvd':
        return 'Album'

    torrent_metadata = tp.TorrentFileParser(jps_torrent_object).parse()
    music_extensions = ['.flac', '.mp3', '.ogg', '.alac', '.m4a', '.wav', '.wma', '.ra']
    off_vocal_phrases = ['off-vocal', 'offvocal', 'off vocal', 'inst.', 'instrumental', 'english ver', 'japanese ver', 'korean ver']
    track_count = 0
    has_cue = False
    track_extensions = set()
    for file in torrent_metadata['info']['files']:
        file_path = file['path'][-1].lower()

        if file_path.endswith('.iso'):
            return 'Album'

        if file_path.endswith('.cue'):
            has_cue = True

        if list(filter(file_path.endswith, music_extensions)) and \
                not any(substring in file_path for substring in off_vocal_phrases):
            #  Count music files which are not an off-vocal or instrumental
            logger.debug(f"Deciding if EP with torrent with these tracks: {file['path'][-1]}")
            track_count += 1
            track_extensions.add(file_path.split('.')[-1])

    if has_cue and track_extensions == {'flac'}:
        logger.debug(f'Upload is not an EP as it has a .cue file and only .flac files')
        return 'Album'

    if track_count < 7:
        logger.debug(f'Upload is an EP as it has {track_count} standard tracks')
        return 'EP'
    else:
        logger.debug(f'Upload is not an EP as it has {track_count} tracks')
        return 'Album'



from jps2sm.utils import GetConfig

# External packages
from requestsloginsession import RequestsLoginSession
from loguru import logger


def jpopsuki(url, test_login=False):
    """
    Get content from JPS

    :param url:
    :param test_login: Disable test login
    :return: data
    """
    config = GetConfig()

    jps_login_url = "https://jpopsuki.eu/login.php"
    jps_test_url = "https://jpopsuki.eu"
    jps_success = '<div id="extra1"><span></span></div>'
    login_data = {'username': config.jps_user, 'password': config.jps_pass}

    jps_session = RequestsLoginSession(jps_login_url, login_data, jps_test_url, jps_success, test_login)

    return jps_session.retrieve_content(url)


def sugoimusic(url, method="get", post_data=None, post_data_files=None, test_login=False):
    """
    Get/Post content to SM

    :param post_data_files: Files to send in POST
    :param post_data: Parameters to send in POST
    :param method: HTML method
    :param url: URL to parse
    :param test_login: Disable test login
    :return: data
    """

    config = GetConfig()

    sm_login_url = "https://sugoimusic.me/login.php"
    sm_test_url = "https://sugoimusic.me/"
    sm_success = "Enabled users"
    login_data = {'username': config.sm_user, 'password': config.sm_pass}

    sm_session = RequestsLoginSession(sm_login_url, login_data, sm_test_url, sm_success, test_login)

    return sm_session.retrieve_content(url, method, post_data, post_data_files)




class VideoOptions:
    """
    Store Video option constants
    """

    VideoMedias = ('DVD', 'Blu-Ray', 'VHS', 'VCD', 'TV', 'HDTV', 'WEB')
    badcontainers = ('ISO', 'VOB', 'MPEG', 'AVI', 'MKV', 'WMV', 'MP4')
    badcodecs = ('MPEG2', 'h264')
    badformats = badcontainers + badcodecs
    resolutions = ('720p', '1080i', '1080p')


class Categories:
    """
    Store category constants
    """

    # Store JPS to SM Category translation, defines which JPS Cat gets uploaded to which SM Cat
    # key: JPS category name
    # value: SM category ID
    JPStoSM = {
        'Album': 0,
        'EP': 1,  # Does not exist on JPS
        'Single': 2,
        'Bluray': 3,  # Does not exist on JPS
        'DVD': 4,
        'PV': 5,
        'Music Performance': 6,  # Does not exist on JPS
        'TV-Music': 7,  # Music Show
        'TV-Variety': 8,  # Talk Show
        'TV-Drama': 9,  # TV Drama
        'Pictures': 10,
        'Misc': 11,
    }

    SM = {
        'Album': 0,
        'EP': 1,  # Does not exist on JPS
        'Single': 2,
        'Bluray': 3,  # Does not exist on JPS
        'DVD': 4,
        'PV': 5,
        'Music Performance': 6,  # Does not exist on JPS
        'TV Music': 7,  # TV-Music
        'TV Variety': 8,  # TV-Variety
        'TV Drama': 9,  # TV-Drama
        'Pictures': 10,
        'Misc': 11,
    }

    JPS = [
        'Album', 'Single', 'PV', 'DVD', 'TV-Music', 'TV-Variety', 'TV-Drama', 'Fansubs', 'Pictures', 'Misc'
    ]

    Video = ('Bluray', 'DVD', 'PV', 'TV-Music', 'TV-Variety', 'TV-Drama', 'Music Performance', 'Fansubs')

    Music = ('Album', 'Single')

    # JPS Categories where release date cannot be entered and therefore need to be processed differently
    NonDate = ('TV-Music', 'TV-Variety', 'TV-Drama', 'Fansubs', 'Pictures', 'Misc')
    # JPS Categories where no release data is present and therefore need to be processed differently
    NonReleaseData = ('Pictures', 'Misc')
    # Music and Music Video Torrents, for category validation. This must match the cateogry headers in JPS for an artist, hence they are in plural
    NonTVCategories = ('Albums', 'Singles', 'DVDs', 'PVs')
    # Categories that should have some of their mediainfo stripped if present, must match indices in Categories.SM
    SM_StripAllMediainfo = (0, 1, 2, 11)  # Album, EP, Single, Misc - useful to have duration if we have it added to the description
    SM_StripAllMediainfoExcResolution = 10  # Pictures - useful to have resolution if we have it


class JPSTorrentView:
    """
    Store the JPS logic of sorting torrents - used by get_batch_jps_group_torrent_ids()
    """

    sort_by = {
        'name': 's1',
        'year': 's2',
        'time': 's3',  # snatched time for snatched, seeding time for seeding, added for uploaded and recent
        'size': 's4',
        'snatches': 's5',
        'seeders':  's6',
        'leechers': 's7'
    }



import os
from io import BytesIO
from typing import List, Tuple, Dict, Union

# Third-party modules
from pymediainfo import MediaInfo
import torrent_parser as tp
from pyunpack import Archive
from pathlib import Path
import tempfile

from loguru import logger


def get_mediainfo(jps_torrent_object: BytesIO, media: str, media_roots: List[str]) -> Tuple[str, Dict[str, str]]:
    """
    Get filename(s) of video files in the torrent and run mediainfo and capture the output, extract if DVD found (Blurays not yet supported)
    then set the appropriate fields for the upload

    :param jps_torrent_object: bytes: BytesIO object of the JPS torrent
    :param media: str Validated media from collate()
    :param media_roots: Sanitised MediaDirectories from cfg for use by get_media_location()
    :return: mediainfo, releasedataout
            mediainfo: Mediainfo text output of the file(s)
            releaseadtaout: Fields gathered from mediainfo for SM upload
    """

    def validate_container(file_extension: str) -> str:
        """
        Map known 'bad' / alternative file extensions of containers to the primary name of the container type.
        """
        extensions = {
            "TP": "TS",
            "TSV": "TS",
            "TSA": "TS",
            "M4V": "MP4",
            "MPG": "MPEG",
        }

        validated_extension = file_extension  # Default is to not validate if it is not in the known incorrect list above - let SM catch it.
        for old, new in extensions.items():
            if file_extension.upper() == old:
                validated_extension = new

        return validated_extension

    def validate_codec(codec: str) -> str:
        """
        Map known alternative names for codecs returned by mediainfo to the primary name of the codec
        """
        codecs = {
            "MPEG Video": "MPEG-2",
            "AVC": "h264",
            "HEVC": "h265",
            "MPEG-4 Visual": "DivX",  # MPEG-4 Part 2 / h263 , usually xvid / divx
            "VP09": "VP9",
            "VP08": "VP8",
        }

        validated_codec = codec  # Default is to not validate it if it is not known in the incorrect list above - let SM catch it.
        for old, new in codecs.items():
            if codec == old:
                validated_codec = new

        return validated_codec

    torrentmetadata = tp.TorrentFileParser(jps_torrent_object).parse()
    torrentname = torrentmetadata['info']['name']  # Directory if >1 file, otherwise it is filename
    # print(torrentmetadata)
    mediainfosall = ""
    releasedataout = {}
    releasedataout['duration'] = 0

    # TODO Need to cleanup the logic to create an overall filename list to parse instead of the 3-way duplication we currently have

    fileforsmfields = None

    if 'files' in torrentmetadata['info'].keys():  # Multiple files
        directory = torrentname
        logger.info(f'According torrent metadata the dir is {directory}')
        file_path = get_media_location(directory, True, media_roots)
        logger.info(f'Path to dir: {file_path}')
        for file in torrentmetadata['info']['files']:
            if len(torrentmetadata['info']['files']) == 1:  # This might never happen, it could be just info.name if so
                filename = os.path.join(*file['path'])
            else:
                releasedataout['multiplefiles'] = True
                filename = os.path.join(*[file_path, *file['path']])  # Each file in the directory of source data for the torrent

            mediainfosall += str(MediaInfo.parse(filename, text=True))
            releasedataout['duration'] += get_mediainfo_duration(filename)  # TODO should this reference file_path directly?
            # Get biggest file and mediainfo on this to set the fields for the release
            maxfile = max(torrentmetadata['info']['files'], key=lambda x: x['length'])  # returns {'length': int, 'path': [str]} of largest file
            fileforsmfields = Path(*[file_path, *maxfile['path']])  # Assume the largest file is the main file that should populate SM upload fields

    else:  # Single file
        releasedataout['multiplefiles'] = False
        filename = torrentname
        file_path = get_media_location(filename, False, media_roots)
        logger.debug(f'Filename for mediainfo: {file_path}')
        mediainfosall += str(MediaInfo.parse(file_path, text=True))
        releasedataout['duration'] += get_mediainfo_duration(file_path)
        fileforsmfields = file_path

    if fileforsmfields is None:
        raise RuntimeError("Error in parsing torrent meta data to get the filename used for populating the SM media fields.")

    if fileforsmfields.suffix == '.iso' and media == 'DVD':
        # If DVD, extract the ISO and run mediainfo against appropriate files, if BR we skip as pyunpack (patool/7z) cannot extract them
        releasedataout['container'] = 'ISO'
        logger.info(f'Extracting ISO {fileforsmfields} to obtain mediainfo on it...')
        isovideoextensions = ('.vob', '.m2ts')
        tempdir = tempfile.TemporaryDirectory()
        Archive(fileforsmfields).extractall(tempdir.name)
        dir_files = []
        for root, subFolder, files in os.walk(tempdir.name):
            for item in files:
                filenamewithpath = os.path.join(root, item)
                dir_files.append(filenamewithpath)
                if list(filter(filenamewithpath.lower().endswith,
                               isovideoextensions)):  # Only gather mediainfo for DVD video files (BR when supported)
                    mediainfosall += str(MediaInfo.parse(filenamewithpath, text=True))
                    releasedataout['duration'] += get_mediainfo_duration(filenamewithpath)

        filesize = lambda f: os.path.getsize(f)
        fileforsmfields = sorted(dir_files, key=filesize)[-1]  # Assume the largest file is the main file that should populate SM upload fields

    # Now we have decided which file will have its mediainfo parsed for SM fields, parse its mediainfo
    mediainforeleasedata = MediaInfo.parse(fileforsmfields)
    # Remove path to file in case it reveals usernames etc.
    replacement = str(Path(file_path).parent)
    mediainfosall = mediainfosall.replace(replacement, '')

    if Path(fileforsmfields).suffix == '.iso' and media == 'DVD':
        tempdir.cleanup()  # TODO This looks like it can be moved to the if block above

    for track in mediainforeleasedata.tracks:
        if track.track_type == 'General':
            # releasedataout['language'] = track.audio_language_list  # Will need to check if this is reliable
            if 'container' not in releasedataout:  # Not an ISO, only set container if we do not already know its an ISO
                releasedataout['container'] = validate_container(track.file_extension.upper())
            else:  # We have ISO - get category data based Mediainfo if we have it
                if track.file_extension.upper() == 'VOB':
                    releasedataout['category'] = 'DVD'
                elif track.file_extension.upper() == 'M2TS':  # Not used yet as we cannot handle Bluray / UDF
                    releasedataout['category'] = 'Bluray'

        if track.track_type == 'Video':
            releasedataout['codec'] = validate_codec(track.format)

            standardresolutions = {
                "3840": "1920",
                "1920": "1080",
                "1280": "720",
                "720": "480",
            }
            for width, height in standardresolutions.items():
                if str(track.width) == width and str(track.height) == height:
                    releasedataout['ressel'] = height

            if 'ressel' in releasedataout.keys():  # Known resolution type, try to determine if interlaced
                if track.scan_type == "Interlaced" or track.scan_type == "MBAFF":
                    releasedataout['ressel'] += "i"
                else:
                    releasedataout['ressel'] += "p"  # Sometimes a Progressive encode has no field set
            else:  # Custom resolution
                releasedataout['ressel'] = 'Other'
                releasedataout['resolution'] = str(track.width) + "x" + str(track.height)

        if track.track_type == 'Audio' or track.track_type == 'Audio #1':  # Handle multiple audio streams, we just get data from the first for now
            if track.format in ["AAC", "DTS", "PCM", "AC3", "Vorbis", "Opus"]:
                releasedataout['audioformat'] = track.format
            elif track.format == "AC-3":
                releasedataout['audioformat'] = "AC3"
            elif track.format == "MPEG Audio" and track.format_profile == "Layer 3":
                releasedataout['audioformat'] = "MP3"
            elif track.format == "MPEG Audio" and track.format_profile == "Layer 2":
                releasedataout['audioformat'] = "MP2"

    logger.debug(f'Mediainfo interpreted data: {releasedataout}')

    return mediainfosall, releasedataout


def get_mediainfo_duration(filename: Union[str, Path]) -> float:
    """
    Get duration in mediainfo for filename

    :param filename:
    :return: float ms
    """
    mediainfo_for_duration = MediaInfo.parse(filename)
    for track in mediainfo_for_duration.tracks:
        if track.track_type == 'General':
            if track.duration is None:
                return 0
            else:
                logger.info(f'Mediainfo duration: {filename} {track.duration}')
                return float(track.duration)  # time in ms


def get_media_location(media_name: str, directory: bool, media_roots: List[str]) -> Path:
    """
    Find the location of the directory or file of the source data for getmediainfo()

    :param media_name: str name of the file or directory
    :param directory: boolean true if dir, false if file
    :param fall_back_file: str fall back search cor
    :param media_roots: Sanitised MediaDirectories from cfg

    :return: full path to file/dir
    """

    # Find the file/dir and stop on the first hit, hopefully OS-side disk cache will mean this will not take too long

    media_location = None
    logger.info(f'Searching for {media_name}...')

    for media_dir_search in media_roots:
        for dirname, dirnames, filenames in os.walk(media_dir_search):
            if directory is True:
                for subdirname in dirnames:
                    if subdirname == media_name:
                        media_location = os.path.join(dirname, subdirname)
                        return Path(media_dir_search, media_location)
            else:
                for filename in filenames:
                    if filename == media_name:
                        media_location = os.path.join(dirname, filename)
                        return Path(media_dir_search, media_location)

    if media_location is None:
        media_not_found_error_msg = f'Mediainfo error - file/directory not found: {media_name} in any of the MediaDirectories specified: {media_roots}'
        logger.error(media_not_found_error_msg)
        raise RuntimeError(media_not_found_error_msg)



import json
import html

from jps2sm.get_data import GetGroupData, GetSMUser, get_torrent_link
from jps2sm.myloginsession import sugoimusic, jpopsuki
from jps2sm.utils import get_valid_filename, GetConfig, HandleCfgOutputDirs
from typing import Type, Tuple

from pathlib import Path
from loguru import logger
import requests

config = GetConfig()
output = HandleCfgOutputDirs(config.directories)


def save_sm_html_debug_output(content: str, torrent_group_data: Type[GetGroupData], jps_torrent_id: str) -> Path:
    """
    Save the html output from uploading the torrent to SM, for debug purposes

    """
    html_debug_output_filename = f"SMuploadresult." + \
                                 get_valid_filename(f"{torrent_group_data.artist[0]}.{torrent_group_data.title}.{torrent_group_data.date}."
                                                    f"JPS_ID{jps_torrent_id}") + \
                                 f".html"
    html_debug_output_path = Path(output.file_dir['html'], html_debug_output_filename)

    with open(html_debug_output_path, "w") as f:
        f.write(content)

    return html_debug_output_path


def download_sm_uploaded_torrents(torrentcount: int, artist: str, title: str) -> None:
    """
    Get last torrentcount torrent DL links that user uploaded using SM API and download them

    :param torrentcount: count of recent torrent links to be downloaded
    :param artist
    :param title

    """
    # TODO Merge this functionality into download_sm_torrent()!

    output_dir = output.file_dir['smtorrents']

    if torrentcount == 0:
        return

    user_recents = sugoimusic(f"https://sugoimusic.me/ajax.php?action=user_recents&limit={torrentcount}")
    user_recents_json = json.loads(user_recents.text)

    smtorrentlinks = {}
    for torrentdata in user_recents_json['response']['uploads']:  # Get list of SM torrent links
        smtorrentlinks[torrentdata['torrentid']] = torrentdata['torrentdl']

    for torrentid, torrentlink in smtorrentlinks.items():
        torrentfile = sugoimusic(torrentlink)
        torrentfilename = get_valid_filename(f'SM {artist} - {title} - {torrentid}.torrent')
        torrentpath = Path(output_dir, torrentfilename)

        with open(torrentpath, "wb") as f:
            f.write(torrentfile.content)
        logger.debug(f'Downloaded SM torrent as {torrentpath}')


def download_sm_torrent(torrent_id: str, artist: str, title: str) -> Path:
    """
    Downloads the SM torrent if it is a dupe, in this scenario we cannot use download_sm_uploaded_torrents() as the user
    has not actually uploaded it.

    :param torrent_id: SM torrentid to be downloaded
    :param artist: SM Artist name, for naming torrent file only
    :param title: SM group title, for naming torrent file only
    :return: name: filename of torrent downloaded
    """
    output_dir = output.file_dir['smtorrents']

    sm_user = GetSMUser()

    file = jpopsuki(
        'https://sugoimusic.me/torrents.php?action='
        f'download&id={torrent_id}&authkey={sm_user.auth_key()}&torrent_pass={sm_user.torrent_password_key()}'
    )
    name = get_valid_filename(
        "SM %s - %s - %s.torrent" % (artist, title, torrent_id)
    )
    sm_torrent_path = Path(output_dir, name)
    with open(sm_torrent_path, "wb") as f:
        f.write(file.content)

    return sm_torrent_path


def get_jps_torrent(jps_torrent_id: str, torrent_group_data: Type[GetGroupData]):
    """
    Download a JPS torrent, ascertaining the link from the torrent_table from the JPS Group

    :param jps_torrent_id: JPS torrent ID
    :param torrent_group_data: Data from the JPS Group
    """

    torrent_link = html.unescape(get_torrent_link(jps_torrent_id, torrent_group_data.torrent_table))
    jps_torrent_file = jpopsuki(f"https://jpopsuki.eu/{torrent_link}")  # Download JPS torrent

    return jps_torrent_file


def download_jps_torrent(jps_torrent_file, torrent_group_data: Type[GetGroupData], release_data):
    """
    Save the JPS torrent

    :param jps_torrent_file: jpopsuki() of the JPS torrent
    :param torrent_group_data: Data from the JPS Group
    :param release_data: Relase data from collate()
    """

    output_dir = output.file_dir['jpstorrents']
    jps_torrent_filename = get_valid_filename(
        f"JPS {torrent_group_data.artist} - {torrent_group_data.title} - {'-'.join(release_data)}.torrent")

    jps_torrent_path = Path(output_dir, jps_torrent_filename)

    with open(jps_torrent_path, "wb") as f:
        f.write(jps_torrent_file.content)



# Standard library packages
import collections
import time
import re
import sys

from jps2sm.get_data import GetGroupData, get_jps_group_data_class
from jps2sm.myloginsession import jpopsuki
from jps2sm.save_data import download_sm_uploaded_torrents
from jps2sm.utils import GetArgs, count_values_dict, GetConfig
from jps2sm.constants import JPSTorrentView

# Third-party packages
from bs4 import BeautifulSoup

from loguru import logger


def batch_mode(mode, user, start=1, end=None, sort=None, order=None):
    """
    Operate batch upload mode
    """
    from jps2sm.jps2sm import collate
    args = GetArgs()

    def batch_stats(final_stats, media_info_mode, dry_run):
        """
        Return statistics from a batch upload.
        If using --recent mode it returns initial statistics after JPS torrents have been scraped, followed by final_stats after the
        SM upload.

        TODO: Needs some cleaning up, but this is better than the stats being completely duplicated as it was before
        """
        if final_stats:
            print('Finished batch upload')
        print(f'--------------------------------------------------------\nOverall stats:'
              f'\nTorrents found at JPS: {batch_uploads_found}'
              f'\nJPS Group data errors: {count_values_dict(batch_uploads_group_errors)}'
              f'\nJPS Groups excluded by user: {len(batch_groups_excluded)}'
              f'\nJPS "V.A." Groups with no contributing artists: {len(batch_groups_va_errors)}'
              )

        if final_stats:
            print(f'Release data (or any other) errors: {count_values_dict(batch_upload_collate_errors)}')
        else:
            print(f'Release data errors: Not yet analysed')

        print(f'Torrents skipped due to max_size filter: {batch_torrent_info["skipped_torrents_max_size"]}'
              f'\nTorrents skipped due to low seeders: {batch_torrent_info["skipped_torrents_low_seeders"]}'  
              f'\nTorrents excluded by user: {batch_torrent_info["skipped_torrents_exc_filter"]}'
              f'\nDuplicates found with torrent hash: {batch_torrent_info["skipped_torrents_duplicate"]}'
              f'\nJPS Torrents downloaded: {batch_torrent_info["jps_torrents_downloaded_count"]}'
              )
        if final_stats:
            if media_info_mode:
                print(f'MediaInfo source data missing: {len(batch_upload_source_data_not_found)}')
            if not dry_run:
                logger.info(f'MediaInfo not submitted errors (use \"--mediainfo\" to fix): {batch_upload_mediainfo_not_submitted}'
                            f'\n\nDuplicates found on SM: {len(batch_upload_dupes_at_upload_smids)}'
                            f'\nNew uploads successfully created: {batch_torrent_info["sm_torrents_uploaded_count"]}'
                            )

    batch_uploads = get_batch_jps_group_torrent_ids(mode=mode, user=user, first=start, last=end, sort=sort, order=order)

    # batch_uploads = { '362613': ['535927'], '354969': ['535926'], '362612': ['535925'], '362611': ['535924'], '181901': ['535923'], '181902': ['535922'] }

    batch_upload_collate_errors = collections.defaultdict(list)
    batch_upload_dupes_at_upload_smids = []
    batch_upload_dupes_at_upload_jpsids = []
    batch_upload_source_data_not_found = []
    batch_upload_mediainfo_not_submitted = 0

    batch_uploads_found = count_values_dict(batch_uploads)

    batch_torrent_info = {
        'skipped_torrents_max_size': 0,
        'skipped_torrents_low_seeders': 0,
        'jps_torrents_downloaded_count': 0,
        'skipped_torrents_exc_filter': 0,
        'sm_torrents_uploaded_count': 0,
        'skipped_torrents_duplicate': 0,
        'dupe_jps_ids': [],
        'dupe_sm_ids': [],
    }

    logger.info(f'Now attempting to upload {batch_uploads_found} torrents.')

    batch_group_data, batch_uploads_group_errors, batch_groups_excluded, batch_groups_va_errors = get_batch_group_data(batch_uploads, args.parsed.exccategory)
    # print(json.dumps(batch_group_data, indent=2))

    max_size = None

    if mode == "recent":
        config = GetConfig()
        max_size = config.max_size_recent_mode

        # Do an initial run of collate() to grab the JPS torrents only so data can be downloaded first
        # TODO Extract release data logic so that collate() does not need to be called twice.
        batch_releases_jps_torrents_downloaded_count = 0

        for jps_group_id, jps_torrent_ids in batch_uploads.items():
            if jps_group_id in batch_uploads_group_errors or jps_group_id in batch_groups_excluded or jps_group_id in batch_groups_va_errors:
                # Skip group if GetGroupData() failed or the group is being excluded by the '-exc' parameter , or if it is a 'V.A.' group and
                # no contrib artists were set
                # TODO Should the jps_group_ids be deleted within get_batch_group_data() ?
                continue

            jps_group_data = get_jps_group_data_class(batch_group_data, jps_group_id)

            try:
                collate_torrent_info_recent = collate(torrentids=jps_torrent_ids, torrentgroupdata=jps_group_data, max_size=max_size, scrape_only=True)
                for collate_result_item, value in collate_torrent_info_recent.items():
                    if isinstance(value, int):
                        batch_torrent_info[collate_result_item] += value
                    elif isinstance(value, list):
                        for item in value:
                            batch_torrent_info[collate_result_item].append(item)
                    else:
                        raise RuntimeError('Expected either int or list in collate_torrent_info.items() from collate()')

            except Exception as exc:
                logger.exception(f'Exception  - {exc} - in collate() on initial run with recent batch mode, however these are skipped'
                                 'as they will be handled in the next run.')
                continue

        batch_stats(final_stats=False, media_info_mode=args.parsed.mediainfo, dry_run=args.parsed.dryrun)

        # Reset stats for the second run of collate()
        # TODO Put batch_stats() into non-nested def and then different dicts can then used instead of reassignment
        batch_torrent_info = {
            'skipped_torrents_max_size': 0,
            'skipped_torrents_low_seeders': 0,
            'jps_torrents_downloaded_count': 0,
            'skipped_torrents_exc_filter': 0,
            'sm_torrents_uploaded_count': 0,
            'skipped_torrents_duplicate': 0,
            'dupe_jps_ids': [],
            'dupe_sm_ids': [],
        }

        input('When these files have been downloaded press enter to continue...')

    for jps_group_id, jps_torrent_ids in batch_uploads.items():
        if jps_group_id in batch_uploads_group_errors or jps_group_id in batch_groups_excluded or jps_group_id in batch_groups_va_errors:
            # Skip group if GetGroupData() failed or the group is being excluded by the '-exc' parameter, or if it is a 'V.A.' group and
            # no contrib artists were set
            # TODO Should the jps_group_ids be deleted within get_batch_group_data() ?
            continue

        jps_group_data = get_jps_group_data_class(batch_group_data, jps_group_id)

        try:
            collate_torrent_info = collate(torrentids=jps_torrent_ids, torrentgroupdata=jps_group_data, max_size=max_size)
            for collate_result_item, value in collate_torrent_info.items():
                if isinstance(value, int):
                    batch_torrent_info[collate_result_item] += value
                elif isinstance(value, list):
                    for item in value:
                        batch_torrent_info[collate_result_item].append(item)
                else:
                    raise RuntimeError('Expected either int or list in collate_torrent_info.items() from collate()')

            if not args.parsed.dryrun:
                download_sm_uploaded_torrents(collate_torrent_info['sm_torrents_uploaded_count'], jps_group_data.artist, jps_group_data.title)
        except KeyboardInterrupt:  # Allow Ctrl-C to exit without showing the error multiple times and polluting the final error dict
            break  # Still continue to get error dicts and dupe list so far
        except Exception as exc:
            # TODO These should all be custom exceptions
            if str(exc).startswith('The exact same torrent file already exists on the site!'):
                sm_dupe_torrentid, jps_dupe_torrentid = re.findall(
                    r'The exact same torrent file already exists on the site! See: https://sugoimusic\.me/torrents\.php\?torrentid=([0-9]+) JPS torrent id\: ([0-9]+)',
                    str(exc))[0]
                batch_upload_dupes_at_upload_smids.append(sm_dupe_torrentid)
                batch_upload_dupes_at_upload_jpsids.append(jps_dupe_torrentid)
            elif str(exc).startswith('Mediainfo error - file/directory not found'):
                # Need to get filename that was not found
                missing_file = re.findall(r'Mediainfo error - file/directory not found: (.+) in any of the MediaDirectories', str(exc))
                batch_upload_source_data_not_found.append(missing_file)
            elif str(exc).startswith('You do not appear to have entered any MediaInfo data for your video upload.'):
                batch_upload_mediainfo_not_submitted += 1
            else:
                # Catch all for any exception
                logger.exception(
                    f'Error with collating/retrieving release data for {jps_group_id} torrentid(s) {",".join(jps_torrent_ids)}, skipping upload')
                batch_upload_collate_errors[jps_group_id] = jps_torrent_ids
            continue

    if batch_uploads_group_errors:
        logger.error('The following JPS groupid(s) (torrentid(s) shown for reference) had errors in retrieving group data, '
                     'keep this data safe and you can possibly retry with it in a later version:')
        logger.error(batch_uploads_group_errors)
        logger.error(f'Total: {count_values_dict(batch_uploads_group_errors)}')
    if batch_groups_excluded:
        logger.info('The following groups were excluded due to user-specified filters:')
        logger.info(f'{batch_groups_excluded}\nTotal: {len(batch_groups_excluded)}')
    if batch_groups_va_errors:
        logger.warning('The following JPS groupid(s) were "bad V.A." groups - they are Various Artist groups with the artist set as "V.A." and'
                       ' the contributing artists should be set as the actual list of artists, however these are missing:')
        logger.warning(batch_groups_va_errors)
        logger.warning(f'Total: {len(batch_groups_va_errors)}')
    if batch_upload_collate_errors:
        logger.error('The following JPS groupid(s) and corresponding torrentid(s) had errors either in collating/retrieving '
                     'release data or in performing the actual upload to SM (although group data was retrieved OK), '
                     'keep this data safe and you can possibly retry with it in a later version:')
        logger.error(batch_upload_collate_errors)
        logger.error(f'Total: {count_values_dict(batch_upload_collate_errors)}')
    if batch_upload_dupes_at_upload_smids:  # Dupes found with dupe error from SM
        logger.warning(
            'The following torrents have already been uploaded to the site, and were found during the SM upload, the SM torrents were downloaded so you can cross seed:')
        logger.warning(f'SM duplicate torrent ids: {batch_upload_dupes_at_upload_smids}\nJPS duplicate torrent ids: {batch_upload_dupes_at_upload_jpsids}')
        logger.warning(f'Total: {len(batch_upload_dupes_at_upload_smids)}')
    if batch_torrent_info['dupe_sm_ids']:  # Dupes found by decide_duplicate()
        logger.warning('The following torrents have already been uploaded to the site, and were found by searching for the torrent hash on SM, the SM torrents were download so you can cross seed:')
        logger.warning(f'SM duplicate torrent ids: {batch_torrent_info["dupe_sm_ids"]}\nPS duplicate torrent ids: {batch_torrent_info["dupe_jps_ids"]}'
                       f'\nTotal: {len(batch_torrent_info["dupe_sm_ids"])}')
    if batch_upload_source_data_not_found:
        logger.error('The following file(s)/dir(s) were not found in your MediaDirectories specified in jps2sm.cfg and the upload was skipped:')
        logger.error(batch_upload_source_data_not_found)
        logger.error(f'Total: {len(batch_upload_source_data_not_found)}')

    batch_stats(final_stats=True, media_info_mode=args.parsed.mediainfo, dry_run=args.parsed.dryrun)


def get_batch_jps_group_torrent_ids(mode, user, first=1, last=None, sort=None, order=None):
    """
    Iterates through pages of uploads on JPS and gathers the jps_group_ids and corresponding jps_torrent_id and returns
    a dict in the format of jps_group_id: [jps_torrent_id]

    :param mode: Area to get batch torrent ids from:
        'uploaded' for a user's uploads,
        'seeding' for the user's currently seeding torrents,
        'snatched' for a user's snatched torrents and
        'recent' for uploads recently uploaded to JPS from ANY user,
    :param user: JPS userid, defined by --batchuser or the SM user specified in jps2sm.cfg
    :param first: upload page number to start at
    :param last: upload page to finish at
    :param sort: Sort the JPS torrents page by a specific column, one of: {",".join(JPSTorrentView.sort_by.keys())}
    :param order: Order by ASC or DESC
    :return: batch_uploads: dict
    """

    # sort_by = {
    #    'name': 's1',
    #    'year': 's2',
    #    'time': 's3',  # snatched time for snatched, seeding time for seeding, added for uploaded and recent
    #    'size': 's4',
    #    'snatches': 's5',
    #    'seeders':  's6',
    #    'leechers': 's7'
    # }

    def get_sort_mode(user_sort):
        # By default:
        # uploaded mode: sort by uploaded time ASC
        # snatched mode: sort by uploaded time ASC
        # seeding mode: sort by name ASC - it is pointless to sort by seeded time as this random on whenever your torrent client
        #   last announced to the tracker
        # recent mode: sort by uploaded DESC - to upload all items recent uploaded to JPS

        default_sort_order_init = "ASC"
        if user_sort is not None:
            return JPSTorrentView.sort_by[user_sort], default_sort_order_init
        if mode == 'snatched' or mode == 'uploaded':
            return JPSTorrentView.sort_by['time'], default_sort_order_init
        elif mode == 'seeding':
            return JPSTorrentView.sort_by['name'], default_sort_order_init
        elif mode == 'recent':
            default_sort_order_init = "DESC"
            return JPSTorrentView.sort_by['time'], default_sort_order_init

    sort_mode, default_sort_order = get_sort_mode(sort)

    if order is not None:
        order_way = str(order).upper()  # Use --batchsortorder specified by user if present
    else:
        order_way = default_sort_order  # Else use sensible default defined in get_sort_mode()

    if not last and mode != 'recent':
        # Ascertain last page if not provided for seeding and snatched modes

        res = jpopsuki(f"https://jpopsuki.eu/torrents.php?type={mode}&userid={user}")
        soup = BeautifulSoup(res.text, 'html5lib')
        linkbox = str(soup.select('#content #ajax_torrents .linkbox')[0])
        try:
            last = re.findall(
                fr'page=([0-9]*)&amp;order_by=s3&amp;order_way=DESC&amp;type={mode}&amp;userid=(?:[0-9]*)&amp;disablegrouping=1(?:\'\);|&amp;action=advanced)"><strong> Last &gt;&gt;</strong>',
                linkbox)[0]
        except IndexError:
            # There is only 1 page of uploads if the 'Last >>' link cannot be found
            last = 1
    elif not last and mode == 'recent':
        # We do not need to ascertain the last page for recent mode as we never use this - we are not trying to jps2sm
        # *every* torrent!
        last = 1

    # Just for debugging purposes - get the jps_sort_name from the key when specifying the value in JPSTorrentView.sort_by{}
    jps_sort_name = list(JPSTorrentView.sort_by.keys())[list(JPSTorrentView.sort_by.values()).index(sort_mode)]
    if not first:
        first = 1
    logger.debug(f'Batch user is {user}, batch mode is {mode}, '
                 f'sort is {sort}, JPS sort column is {sort_mode} - {jps_sort_name}, '
                 f'order by is {order_way} '
                 f'first page is {first}, last page is {last}')

    batch_uploads = collections.defaultdict(list)

    # Parse every torrent page and add to dict
    for i in range(first, int(last) + 1):
        if mode == 'snatched' or mode == 'uploaded' or mode == 'seeding':
            batch_upload_url = f"https://jpopsuki.eu/torrents.php?page={i}&order_by={sort_mode}&order_way={order_way}&type={mode}&userid={user}&disablegrouping=1"
        elif mode == 'recent':
            batch_upload_url = f"https://jpopsuki.eu/torrents.php?page={i}&order_by={sort_mode}&order_way={order_way}&disablegrouping=1"
        else:
            raise RuntimeError("Unknown batch mode set")

        batch_upload_page = jpopsuki(batch_upload_url)
        logger.info(batch_upload_url)
        # print batch_upload_page.text
        soup2 = BeautifulSoup(batch_upload_page.text, 'html5lib')
        try:
            torrent_table = str(soup2.select('#content #ajax_torrents .torrent_table tbody')[0])
        except IndexError:
            quota_exceeded = re.findall('<title>Browse quota exceeded :: JPopsuki 2.0</title>', batch_upload_page.text)
            if quota_exceeded:
                logger.error('Browse quota exceeded :: JPopsuki 2.0')
                sys.exit(1)
            else:
                raise
        # Find all jps_group_id/jps_torrent_id pairs and returns a list of tuples
        all_jps_group_ids_and_torrent_ids = re.findall('torrents.php\?id=([0-9]+)&amp;torrentid=([0-9]+)', torrent_table)
        logger.info(f'jps_group_ids and jps_torrent_ids found on page: {all_jps_group_ids_and_torrent_ids}')

        for jps_group_id, jps_torrent_id in all_jps_group_ids_and_torrent_ids:
            batch_uploads[jps_group_id].append(jps_torrent_id)
        time.sleep(5)  # Sleep as otherwise we hit JPS browse quota

    logger.debug(f'jps_group_ids and jps_torrent_ids found on all pages: {batch_uploads}')
    return batch_uploads


def get_batch_group_data(batch_uploads, excluded_category):
    """
    Iterate through batch_uploads and run GetGroupData on each group and store the data in batch_group_data{}

    :param batch_uploads: dict, Contains jps_group_id: jps_torrent_id of all uploads
    :param excluded_category: str, JPS Category name to be excluded
    :return batch_group_data: dict, multi dimensional dict contain group data of all uploads
    :return batch_group_errors: list, All JPS jps_group_ids where GetGroupData failed
    :return batch_groups_excluded: list, All jps_group_ids that were excluded by the user, currently --exccategory
    :return batch_groups_va_errors: list, jps_group_ids that were 'V.A.' torrents with no contrib artists set
    """

    batch_group_errors = collections.defaultdict(list)
    batch_groups_excluded = []
    batch_groups_va_errors = []
    batch_group_data = {}

    for jps_group_id, jps_torrent_id in batch_uploads.items():
        try:
            logger.info('-------------------------')
            group_data_output = GetGroupData("https://jpopsuki.eu/torrents.php?id=%s" % jps_group_id)
            batch_group_data[jps_group_id] = {}
            batch_group_data[jps_group_id] = group_data_output.all()

            if batch_group_data[jps_group_id]["category"] == excluded_category:
                logger.debug(
                    f'Excluding jps_group_id {jps_group_id} as it is {batch_group_data[jps_group_id]["category"]} group and these are being skipped')
                batch_groups_excluded.append(jps_group_id)
                continue
        except KeyboardInterrupt:  # Allow Ctrl-C to exit without showing the error multiple times and polluting the final error dict
            break  # Still continue to get error dicts and dupe list so far
        except Exception as exc:
            # Catch all for any exception
            va_no_contrib_artists_error = "V.A. torrent with to contrib artists set - torrent has no valid artists so this cannot be uploaded."
            if str(exc) == va_no_contrib_artists_error:
                logger.error(va_no_contrib_artists_error)
                batch_groups_va_errors.append(jps_group_id)
                continue
            logger.exception(
                'Error with retrieving group data for jps_group_id %s jps_torrent_id %s, skipping upload' % (jps_group_id, ",".join(jps_torrent_id)))
            batch_group_errors[jps_group_id] = jps_torrent_id
            continue

    return batch_group_data, batch_group_errors, batch_groups_excluded, batch_groups_va_errors
